#pragma once

#include <cstdint>
#include "pdk.h"
#include "symbols.h"

class CPadaukDisasm
{
private:
    static bool opcode_table_patched;

protected:
    struct TAR {
        char const args[4];
        char const *pargs;
        char const * (CPadaukDisasm::*make_arg)(void);
        struct {
            char const *bits;
            unsigned const mask;
        } e[4];
    } static args[];

    struct TOP {
        char const * const mnemonic;
        union {
            char const * const s;
            TAR const *p;
        } args;
        char const * const flags;
        uint16_t const opcode[4];
        char const * const description;
    } static ops[];

    char sa[48];
    int8_t oi[UINT16_MAX + 1];
    unsigned opcode;

    virtual unsigned get_m(void) = 0;
    virtual unsigned get_w(void) = 0;
    virtual unsigned get_p(void) = 0;
    virtual unsigned get_b(void) = 0;
    virtual void     get_b_m(unsigned &b, unsigned &m) = 0;
    virtual unsigned get_j(void) = 0;
    virtual unsigned get_f(void) { return 0; }
    virtual unsigned get_d(void) { return 0; }

    virtual char const *  a_(void) { return ""; }
    virtual char const *  aa(void) { return "a"; }
    virtual char const *  ak(void);
    virtual char const * aak(void);
    virtual char const * aam(void);
    virtual char const * ama(void);
    virtual char const *  am(void);
    virtual char const *  aw(void);
    virtual char const * awa(void);
    virtual char const * aaw(void);
    virtual char const * aap(void);
    virtual char const * apa(void);
    virtual char const * apb(void);
    virtual char const * amb(void);
    virtual char const *  aj(void);
    virtual char const *  af(void);
    virtual char const *  ad(void);

    static TAR const *find_arg(char const * const s);
    static void patch_opcode_table(void);
    void init_disasm(unsigned const e);
    static char const *instruction_encoding(unsigned const size, unsigned const opcode, char const * const bits);
    static void print_instruction_set(unsigned const e, unsigned const b);

public:
    CSymbols io_symbols;
    CSymbols mem_symbols;

    CPadaukDisasm() { io_symbols.default_symbol.assign("io."); mem_symbols.default_symbol.assign("M"); patch_opcode_table(); }
    static CPadaukDisasm * New(unsigned bits);
    void disasm(uint16_t const w);
    void disasm(CPDK &pdk);
    static void compare_opcodes(void);
};

class CPadaukDisasm16 : public CPadaukDisasm
{
protected:
    virtual unsigned get_m(void) { return opcode & 0x01FF; }
    virtual unsigned get_w(void) { return opcode & 0x01FE; };
    virtual unsigned get_p(void) { return opcode & 0x003F; };
    virtual unsigned get_b(void) { return (opcode & 0x01C0) >> 6; }
    virtual void get_b_m(unsigned &b, unsigned &m) { b = (opcode & 0x0E00) >> 9; m = opcode & 0x01FF; }
    virtual unsigned get_j(void) { return opcode & 0x1FFF; };
    virtual unsigned get_f(void) { return opcode & 0x000F; };
    virtual unsigned get_d(void) { return opcode & 0x001F; };
public:
    CPadaukDisasm16() { init_disasm(0); }
    static void print_instruction_set(void) { CPadaukDisasm::print_instruction_set(0, 16); }
};

class CPadaukDisasm15 : public CPadaukDisasm
{
protected:
    virtual unsigned get_m(void) { return opcode & 0x00FF; }
    virtual unsigned get_w(void) { return opcode & 0x00FE; };
    virtual unsigned get_p(void) { return opcode & 0x007F; };
    virtual unsigned get_b(void) { return (opcode & 0x0380) >> 7; }
    virtual void get_b_m(unsigned &b, unsigned &m) { b = (opcode & 0x0380) >> 7; m = opcode & 0x007F; }
    virtual unsigned get_j(void) { return opcode & 0x0FFF; };
public:
    CPadaukDisasm15() { init_disasm(1); }
    static void print_instruction_set(void) { CPadaukDisasm::print_instruction_set(1, 15); }
};

class CPadaukDisasm14 : public CPadaukDisasm
{
protected:
    virtual unsigned get_m(void) { return opcode & 0x007F; }
    virtual unsigned get_w(void) { return opcode & 0x007E; };
    virtual unsigned get_p(void) { return opcode & 0x003F; };
    virtual unsigned get_b(void) { return (opcode & 0x01C0) >> 6; }
    virtual void get_b_m(unsigned &b, unsigned &m) { b = (opcode & 0x01C0) >> 6; m = opcode & 0x003F; }
    virtual unsigned get_j(void) { return opcode & 0x07FF; };
public:
    CPadaukDisasm14() { init_disasm(2); }
    static void print_instruction_set(void) { CPadaukDisasm::print_instruction_set(2, 14); }
};

class CPadaukDisasm13 : public CPadaukDisasm
{
protected:
    virtual unsigned get_m(void) { return opcode & 0x003F; }
    virtual unsigned get_w(void) { return opcode & 0x001E; };
    virtual unsigned get_p(void) { return opcode & 0x001F; };
    virtual unsigned get_b(void) { return (opcode & 0x00E0) >> 5; }
    virtual void get_b_m(unsigned &b, unsigned &m) { b = (opcode & 0x00E0) >> 5; m = opcode & 0x000F; }
    virtual unsigned get_j(void) { return opcode & 0x03FF; };
public:
    CPadaukDisasm13() { init_disasm(3); }
    static void print_instruction_set(void) { CPadaukDisasm::print_instruction_set(3, 13); }
};
